/* $Id$

This file is part of NetVis
Copyright @ 2008 Mikko Rasa, Mikkosoft Productions
Distributed unter the GPL
*/

#include <limits>
#include <cstdlib>
#include <cmath>
#include <signal.h>
#include <netinet/ether.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <msp/core/except.h>
#include <msp/debug/profilingscope.h>
#include <msp/gl/blend.h>
#include <msp/gl/framebuffer.h>
#include <msp/gl/immediate.h>
#include <msp/gl/matrix.h>
#include <msp/gl/misc.h>
#include <msp/gl/projection.h>
#include <msp/gl/texture2d.h>
#include <msp/gl/transform.h>
#include <msp/strings/formatter.h>
#include <msp/time/units.h>
#include <msp/time/utils.h>
#include "history.h"
#include "host.h"
#include "netvis.h"
#include "packet.h"
#include "port.h"
#include "resolver.h"

using namespace std;
using namespace Msp;

NetVis::NetVis(int argc, char **argv):
	pcap(0),
	resolver(0),
	wnd(0),
	font(0),
	max_hosts(1000),
	max_visible_hosts(30),
	frames(0)
{
	if(argc<2)
		throw UsageError("No interface given");
	iface = argv[1];

	char err[1024];
	pcap = pcap_open_live(iface.c_str(), 128, true, 0, err);
	if(!pcap)
		throw Exception(err);

	if(pcap_setnonblock(pcap, true, err)==-1)
		throw Exception(err);

	pcap_lookupnet(iface.c_str(), &localnet, &localnet_mask, err);
	localnet = ntohl(localnet);
	localnet_mask = ntohl(localnet_mask);

	resolver = new Resolver;

	wnd = new Graphics::SimpleGLWindow(1024, 768);
	wnd->set_title("NetVis");
	wnd->signal_close.connect(sigc::bind(sigc::mem_fun(this, &NetVis::exit), 0));
	wnd->show();

	GL::enable(GL::BLEND);
	GL::blend_func(GL::SRC_ALPHA, GL::ONE_MINUS_SRC_ALPHA);

	font = new GL::Font;
	DataFile::load(*font, "dejavu-10.font");

	history = new History(*this, 301, 100);

	catch_signal(SIGINT);
}

NetVis::~NetVis()
{
	delete resolver;

	delete font;
	delete wnd;

	pcap_close(pcap);
	for(map<unsigned, Host *>::iterator i=hosts.begin(); i!=hosts.end(); ++i)
		delete i->second;
	for(map<unsigned, Host *>::iterator i=disabled_hosts.begin(); i!=disabled_hosts.end(); ++i)
		delete i->second;
	for(map<unsigned, Port *>::iterator i=ports.begin(); i!=ports.end(); ++i)
		delete i->second;
	for(list<Packet *>::iterator i=packets.begin(); i!=packets.end(); ++i)
		delete *i;
}

void NetVis::tick()
{
	Msp::Time::TimeStamp t = Msp::Time::now();
	Msp::Time::TimeDelta dt;
	if(tick_t)
		dt = t-tick_t;
	tick_t = t;

	if(tick_t>fps_t+Msp::Time::sec)
	{
		fps = frames/((tick_t-fps_t)/Msp::Time::sec);
		fps_t = tick_t;
		frames = 0;
	}

	wnd->get_display().tick();

	while(pcap_dispatch(pcap, -1, &capture_handler, reinterpret_cast<unsigned char *>(this))>0) ;

	resolver->tick();
	history->tick(tick_t);


	float min_activity = numeric_limits<float>::max();
	for(map<unsigned, Host *>::iterator i=hosts.begin(); i!=hosts.end(); ++i)
	{
		i->second->tick(dt);
		min_activity = min(min_activity, i->second->get_activity());
	}
	float del_limit = pow(10, 6-0.1*(max_hosts-hosts.size()-disabled_hosts.size()));
	for(map<unsigned, Host *>::iterator i=disabled_hosts.begin(); i!=disabled_hosts.end();)
	{
		i->second->tick(dt);

		if(i->second->get_activity()>min_activity)
		{
			i->second->set_active(true);
			hosts.insert(*i);
			disabled_hosts.erase(i++);
		}
		else if(i->second->get_activity()<del_limit)
		{
			delete i->second;
			disabled_hosts.erase(i++);
		}
		else
			++i;
	}

	if(hosts.size()>max_visible_hosts)
	{
		list<float> activity;
		for(map<unsigned, Host *>::iterator i=hosts.begin(); i!=hosts.end(); ++i)
			activity.push_back(i->second->get_activity());
		activity.sort();

		list<float>::iterator j = activity.begin();
		advance(j, activity.size()-max_visible_hosts);
		float limit = *j;

		for(map<unsigned, Host *>::iterator i=hosts.begin(); i!=hosts.end();)
		{
			if(i->second->get_activity()<limit)
			{
				i->second->set_active(false);
				disabled_hosts.insert(*i);
				hosts.erase(i++);
			}
			else
				++i;
		}
	}

	for(list<Packet *>::iterator i=packets.begin(); i!=packets.end();)
	{
		(*i)->tick(dt);
		if((*i)->get_stale())
		{
			delete *i;
			i = packets.erase(i);
		}
		else
			++i;
	}

	render();
	wnd->swap_buffers();

	++frames;
}

void NetVis::render()
{
	GL::clear(GL::COLOR_BUFFER_BIT);

	GL::matrix_mode(GL::PROJECTION);
	GL::load_identity();
	GL::ortho_centered(1024, 768);
	GL::matrix_mode(GL::MODELVIEW);
	GL::load_identity();

	for(map<unsigned, Host *>::iterator i=hosts.begin(); i!=hosts.end(); ++i)
		i->second->render();
	{
		GL::Immediate imm((GL::COLOR4_UBYTE, GL::VERTEX2));
		imm.begin(GL::QUADS);
		for(list<Packet *>::iterator i=packets.begin(); i!=packets.end(); ++i)
			(*i)->render(imm);
		imm.end();
	}

	GL::push_matrix();
	GL::translate(-500, 360, 0);
	unsigned n = 0;
	for(map<unsigned, Port *>::iterator i=ports.begin(); (i!=ports.end() && n<50); ++i, ++n)
	{
		i->second->render();
		GL::translate(0, -12, 0);
	}
	GL::pop_matrix();

	GL::push_matrix();
	GL::translate(-500, -360, 0);
	GL::scale_uniform(10);
	font->draw_string(format("%d hosts", hosts.size()+disabled_hosts.size()));
	GL::translate(0, -1.2, 0);
	font->draw_string(format("%.2f fps", fps));
	GL::Texture::unbind();
	GL::pop_matrix();

	GL::push_matrix();
	GL::translate(170, -370, 0);
	history->render();
	GL::pop_matrix();
}

Host &NetVis::get_host(unsigned a)
{
	map<unsigned, Host *>::iterator i = hosts.find(a);
	if(i!=hosts.end())
		return *i->second;

	i = disabled_hosts.find(a);
	if(i!=disabled_hosts.end())
		return *i->second;

	Host *host = new Host(*this, a);
	if((a&localnet_mask)==localnet)
		host->set_local(true);
	resolver->push(host);
	host->set_position(Vector2(rand()*400.0/RAND_MAX-200.0, rand()*400.0/RAND_MAX-200.0));
	hosts[a] = host;
	return *host;
}

const Port &NetVis::get_port(unsigned number)
{
	map<unsigned, Port *>::iterator i = ports.find(number);
	if(i!=ports.end())
		return *i->second;
	Port *port = new Port(*this, number);
	ports[number] = port;
	return *port;
}

void NetVis::capture_handler(unsigned char *user, const pcap_pkthdr *, const unsigned char *data)
{
	NetVis *self = reinterpret_cast<NetVis *>(user);

	const ethhdr *eth = reinterpret_cast<const ethhdr *>(data);
	if(ntohs(eth->h_proto)==ETH_P_IP)
	{
		const iphdr *ip = reinterpret_cast<const iphdr *>(eth+1);

		unsigned size = ntohs(ip->tot_len);
		unsigned port = 0;
		if(ip->protocol==IPPROTO_TCP)
		{
			const tcphdr *tcp = reinterpret_cast<const tcphdr *>(ip+1);
			port = min(ntohs(tcp->source), ntohs(tcp->dest));
		}
		else if(ip->protocol==IPPROTO_UDP)
		{
			const udphdr *udp = reinterpret_cast<const udphdr *>(ip+1);
			port = min(ntohs(udp->source), ntohs(udp->dest));
		}
		Host &shost = self->get_host(ntohl(ip->saddr));
		Host *dhost = 0;
		if((ntohl(ip->daddr)&0xFF)!=0xFF)
			dhost = &self->get_host(ntohl(ip->daddr));

		float throttle = shost.send_packet();
		if(throttle<1)
		{
			self->packets.push_back(new Packet(shost, dhost, self->get_port(port).get_color(), size));
			self->packets.back()->tick(-throttle*Msp::Time::sec);
		}

		shost.add_activity(size);
		if(dhost)
			dhost->add_activity(size);

		if((ntohl(ip->saddr)&self->localnet_mask)==self->localnet)
			self->history->activity(0, size);
		else if((ntohl(ip->daddr)&self->localnet_mask)==self->localnet)
			self->history->activity(size, 0);
	}
}

void NetVis::sighandler(int)
{
	exit(0);
}

Application::RegApp<NetVis> NetVis::reg;
